/**
 * Advanced Internet Service Project
 * 
 * Speaker recognition enhanced voice conference.
 * 
 * Authors: Yancheng Li, Liang Wei, Zhaoyuan Zhang
 * 
 */

import java.io.DataOutputStream;
import java.io.IOException;

import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;

public class AudioClient extends Thread {
	/**
	 * Flag for debugging messages. If true, some messages are dumped to the
	 * console during operation.
	 */
	private TargetDataLine m_targetLine;
	private SourceDataLine m_sourceLine;
	private volatile boolean m_bRecording;
	private volatile boolean m_bTraining;
	private int trainingOffset; 
	private int m_nExternalBufferSize;
	private DataOutputStream out;
	private ClientInterface clientInterface;
	private double[] data;
	private int dataOffset;
	private AudioFormat audioFormat;
	private byte[] trainingData;

	/*
	 * We have to pass an AudioFormat to describe in which format the audio data
	 * should be recorded and played.
	 */
	public AudioClient(AudioFormat format, int nInternalBufferSize,
			int nExternalBufferSize, String strMixerName, DataOutputStream out)
			throws LineUnavailableException {
		this.out = out;
		this.audioFormat = format;
		trainingData = new byte[AudioUtil.TRAINING_FRAME_SIZE];
		m_bTraining = false;
		trainingOffset = 0;
		
		/*
		 * We retrieve and open the recording and the playback line.
		 */
		DataLine.Info targetInfo = new DataLine.Info(TargetDataLine.class,
				format, nInternalBufferSize);
		DataLine.Info sourceInfo = new DataLine.Info(SourceDataLine.class,
				format, nInternalBufferSize);
		m_targetLine = (TargetDataLine) AudioSystem.getLine(targetInfo);
		m_sourceLine = (SourceDataLine) AudioSystem.getLine(sourceInfo);
		
		m_targetLine.open(format, nInternalBufferSize);
		m_sourceLine.open(format, nInternalBufferSize);
		m_nExternalBufferSize = nExternalBufferSize;
		
		m_sourceLine.start();
		
		data = new double[AudioUtil.WAVEFORM_SIZE];
		dataOffset = 0;
	}

	public void startRecording() {		
		m_targetLine.start();
		m_bRecording = true;
		
		byte[] abBuffer = new byte[m_nExternalBufferSize];
		for (int i = 0; i < abBuffer.length; i++) {
			abBuffer[i] = AudioUtil.START;
		}
		
		try {
			out.write(abBuffer, 0, m_nExternalBufferSize);					
		} catch (IOException e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}		
	}

	public void stopRecording() {
		m_targetLine.stop();
		m_bRecording = false;
	}
	
	public void startTrainging() {
		m_bTraining = true;
		m_targetLine.start();
		trainingOffset = 0;
		for (int i = 0 ; i < trainingData.length; i++) {
			trainingData[i] = 0;
		}
	}
	public byte[] getTrainingData() {
		m_bTraining = false;
		m_targetLine.stop();
		return trainingData;
	}
	public void run() {
		byte[] abBuffer = new byte[m_nExternalBufferSize];
		int nBufferSize = abBuffer.length;
		@SuppressWarnings("unused")
		int previousLen = 0;
		
		while (true) {

			if (m_bRecording) {
				
				int nBytesRead = m_targetLine.read(abBuffer, 0, nBufferSize);
				try {
					//abBuffer[0] = AudioUtil.DATA;
					out.write(abBuffer, 0, nBytesRead);
					int len = calculateEnergy(abBuffer);
					// printBarLine(len, previousLen);
					previousLen = len;
					
				} catch (IOException e) {
					System.out.println(e.getMessage());
					e.printStackTrace();
				}
			} else if (m_bTraining) {
				int nBytesRead = m_targetLine.read(abBuffer, 0, nBufferSize);
				for (int i = 0; i < nBytesRead; i++) {
					if (trainingOffset < trainingData.length) {
						trainingData[trainingOffset++] = abBuffer[i];
					} else {
						clientInterface.setTitle("Enough samples, ready to save!");
					}
				}
				
				for (int i = 0; i < nBytesRead; i++) {
					if(dataOffset >= AudioUtil.WAVEFORM_SIZE) {
						for (int j = 0; j < AudioUtil.WAVEFORM_SIZE; j++) {
							data[j] = 0;
							dataOffset = 0;
						}
					}
					data[dataOffset++] = abBuffer[i];
				}
				clientInterface.updateWaveFom(data, audioFormat);
			}
			
		}
	}
	
	private boolean isChar(byte data) {
		char charData = (char) data;
		return (charData >= 'a' && charData <= 'z') || (charData >= 'A' && charData <='Z');
	}

	public void write(byte[] buffer, int offset, int len) {
		if (buffer[0] == AudioUtil.NAME && buffer[1] == AudioUtil.NAME && buffer[2] == AudioUtil.NAME) {
			StringBuffer result = new StringBuffer();
			
			for (int i = 3; i < len; i ++) {
				if (isChar(buffer[i])) {
					result.append((char)buffer[i]);
				} else {
					break;
				}
			}
			result.append(" is speaking...");
			System.out.println(result.toString());
			clientInterface.setSpeaker(result.toString());
		} else {//if (!m_bRecording){
			m_sourceLine.write(buffer, offset, len);
			for (int i = 0; i < len; i++) {
				if(dataOffset >= AudioUtil.WAVEFORM_SIZE) {
					for (int j = 0; j < AudioUtil.WAVEFORM_SIZE; j++) {
						data[j] = 0;
						dataOffset = 0;
					}
				}
				data[dataOffset++] = buffer[i];
			}
			clientInterface.updateWaveFom(data, audioFormat);
		}
	}
	
	public int calculateEnergy(byte[] abBuffer) {
		long energy = 0;
		
		for (int i = 1; i < abBuffer.length; i+=2) {
			energy += abBuffer[i] * abBuffer[i]; 
		}
		
		return (int) (energy / 1000);
	}
	
	public void printBarLine(int len, int previousLen) {
		for (int i = 0; i < previousLen + 1; i++) {
			System.out.print('\b');
		}
		
		for (int i = 0; i < previousLen + 1; i++) {
			System.out.print(' ');
		}
		
		for (int i = 0; i < previousLen + 1; i++) {
			System.out.print('\b');
		}
		
		for (int i = 0; i < len; i++) {
			System.out.print('|');
		}
		
		return;
	}
	
	public void setClientInterface(ClientInterface clientInterface) {
		this.clientInterface = clientInterface;
	}
}
