/**
 * Advanced Internet Service Project
 * 
 * Speaker recognition enhanced voice conference. The algorithm comes from 
 * Speaker recognition, Model & Method  Chaohui Wu, Tsinghua University Press.
 * 
 * Authors: Yancheng Li, Liang Wei, Zhaoyuan Zhang
 * 
 */
package SpeakerRegonition;

public class FeatureExtractorUtils {
	
	/* Calculate LPC using Durbin Algorithm */
	
	static double[] calculateLPC(double[] coefficient) {		
		int lpcOrder = 12;
		double[] lpccoef = new double[lpcOrder];
		
		int i, m;
		double temp, alphap;
		double k[] = new double[lpcOrder + 1];
		double a[][] = new double[lpcOrder + 1][lpcOrder + 1];

		k[1] = coefficient[1] / coefficient[0];
		a[1][1] = k[1];
		alphap = coefficient[0] * (1 - k[1] * k[1]);
		for (m = 2; m <= lpcOrder; m++) {
			temp = 0;
			for (i = 1; i < m; i++)
				temp += a[m - 1][i] * coefficient[Math.abs(i - m)];
			k[m] = (coefficient[m] - temp) / alphap;
			a[m][m] = k[m];
			for (i = 1; i < m; i++)
				a[m][i] = a[m - 1][i] - k[m] * a[m - 1][m - i];
			alphap = alphap * (1 - k[m] * k[m]);
		}
		for (i = 0; i < lpcOrder; i++)
			lpccoef[i] = a[lpcOrder][i + 1];
		
		return lpccoef;
	}
	
	/* Calculate auto-correlation */
	public static double[] calculateAutocorrelation(double[] signal) {
		double[] fullr;
		double[] zeror;
		
		final int signalLength = signal.length;
		fullr = new double[2 * signalLength - 1];
		for (int k = 0; k < 2 * signalLength - 1; k++) {
			fullr[k] = 0;
			for (int m = 0; m < 2 * signalLength - k - 1; m++) {
				double a = 0, b = 0;
				if (m > -1 && m < signalLength)
					a = signal[m];
				else
					a = 0;
				if ((m + k - signalLength + 1) > -1
						&& (m + k - signalLength + 1) < signalLength)
					b = signal[m + k - signalLength + 1];
				else
					b = 0;
				fullr[k] = fullr[k] + a * b;
			}
		}

		zeror = new double[signalLength];
		for (int i = 0; i < signalLength; i++)
			zeror[i] = fullr[signalLength - 1 + i];
		
		return zeror;
	}
	
	public static double[][] calculateFrameLPC(double data[][], int frameCount) {
		double[][] frameLPC = new double[frameCount][];

		for (int i = 0; i < frameCount; i++) {
			frameLPC[i] = calculateLPC(calculateAutocorrelation(data[i]));
		}

		return frameLPC;
	}

	public static double[][] calculateFrameLPCC(double lpc[][], int frameCount) {
		double[][] lpcc = new double[frameCount][16];

		for (int i = 0; i < frameCount; i++) {
			lpcc[i][0] = lpc[i][0];
			for (int n = 2; n <= 12; n++) {
				double temp = 0;
				for (int k = 1; k < n; k++)
					temp += (1 - k / n) * lpcc[i][n - 1 - k]
							* lpc[i][k - 1];
				lpcc[i][n - 1] = lpc[i][n - 1] + temp;
			}
			for (int n = 13; n <= 16; n++) {
				for (int k = 1; k <= 12; k++)
					lpcc[i][n - 1] += (1 - k / n)
							* lpcc[i][n - 1 - k] * lpc[i][k - 1];
			}
		}

		return lpcc;
	}

	public static double[] detectPitch(double[][] frameData, int frameCount) {
		double[] pitch;

		pitch = new double[frameCount];

		for (int f = 0; f < frameCount; f++) {

			double[][] pitchData = new double[frameCount][224];
			for (int n = 16; n < 224; n++)
				pitchData[f][n - 16] = frameData[f][n];

			double L, M, M1 = 0, M2 = 0;
			for (int n = 0; n < 100; n++)
				M1 = Math.max(M1, Math.abs(pitchData[f][n]));
			for (int n = 124; n < 224; n++)
				M2 = Math.max(M2, Math.abs(pitchData[f][n]));
			M = Math.min(M1, M2);
			L = 0.68 * M;

			for (int n = 0; n < 224; n++) {
				if (pitchData[f][n] > L)
					pitchData[f][n] = pitchData[f][n] - L;
				else if (pitchData[f][n] < -L)
					pitchData[f][n] = pitchData[f][n] + L;
				else
					pitchData[f][n] = 0;
			}

			double[] coefficient = calculateAutocorrelation(pitchData[f]);

			double maxCoefficient = 0;
			int maxIndex = 0;
			for (int k = 1; k < 124; k++) {
				if (maxCoefficient < Math.abs(coefficient[k])) {
					maxCoefficient = Math.abs(coefficient[k]);
					maxIndex = k;
				}
			}
			
			if (maxCoefficient < 0.25 * Math.abs(coefficient[0]) || maxIndex < 2)
				pitch[f] = 0;
			else
				pitch[f] = 8000 / (maxIndex + 16);
		}

		return pitch;
	}

	public static double[] getOverlappedData(double[] data) {
		int length = data.length - 4;
		double[] result;

		if (length >= 0)
			result = new double[length];
		else
			result = new double[0];

		for (int t = 2; t < data.length - 2; t++) {
			for (int i = -2; i <= 2; i++) {
				double a = 0, b = 0;
				a += i * data[t + i];
				b += i * i;
				result[t - 2] = a / b;
			}
		}

		return result;
	}
}
