/**
 * Advanced Internet Service Project
 * 
 * Speaker recognition enhanced voice conference.
 * 
 * This sampling java file is modified and adapted from musole.sourceforge.net (open source project).
 * 
 * Authors: Yancheng Li, Liang Wei, Zhaoyuan Zhang
 * 
 */

package SpeakerRegonition;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.sound.sampled.*;
import java.util.*;
import java.awt.geom.Line2D;

public class WaveformGraph extends JPanel implements ComponentListener, Runnable {
	/**
	 * 
	 */
	private static final long serialVersionUID = 3806506739281137444L;
	private Thread thread;
	private AudioInputStream wavStream;
	private AudioFormat wavFormat;
	private Vector<Line2D.Double> lines = new Vector<Line2D.Double>();
	private double[] audioDouble;
	private int frames_per_pixel;
	private double[] audioData;
	private boolean initialized;

	public WaveformGraph() {
		setBackground(Color.black);
		addComponentListener(this);
		initialized = false;
	}

	public void componentResized(ComponentEvent e) {
		createWaveForm(audioDouble);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	public void getAudioInputStream(AudioInputStream wavStream) {
		if (wavStream == null) {
			this.wavStream = null;
			wavFormat = null;
			return;
		}
		this.wavStream = wavStream;
		wavFormat = wavStream.getFormat();

	}

	public double[] getAudioData() {
		return audioData;
	}
	
	public void doCreateWaveForm(double[] audioDouble, AudioFormat wavFormat) {
		lines.removeAllElements();
		initialized = true;

		if (audioDouble == null) {
			try {
				audioDouble = new double[(int) (wavStream.getFrameLength() * wavFormat
						.getFrameSize())];
				byte[] tempAudioBytes = new byte[audioDouble.length];
				wavStream.read(tempAudioBytes);
				for (int i = 0; i < audioDouble.length; i++)
					audioDouble[i] = tempAudioBytes[i];
			} catch (Exception ex) {
				return;
			}
		}

		this.audioDouble = audioDouble;

		Dimension d = getSize();
		int w = d.width;
		int h = d.height;
		double[] audioData = null;

		if (wavFormat.getSampleSizeInBits() == 16) {
			int nlengthInSamples = audioDouble.length / 2;
			audioData = new double[nlengthInSamples];

			if (wavFormat.isBigEndian()) {
				for (int i = 0; i < nlengthInSamples; i++) {
					int MSB = (int) audioDouble[2 * i];
					int LSB = (int) audioDouble[2 * i + 1];
					audioData[i] = MSB << 8 | (255 & LSB);
				}
			} else {
				for (int i = 0; i < nlengthInSamples; i++) {
					int LSB = (int) audioDouble[2 * i];
					int MSB = (int) audioDouble[2 * i + 1];
					audioData[i] = MSB << 8 | (255 & LSB);
				}
			}
		} else if (wavFormat.getSampleSizeInBits() == 8) {
			int nlengthInSamples = audioDouble.length;
			audioData = new double[nlengthInSamples];
			if (wavFormat.getEncoding().toString().startsWith("PCM_SIGN")) {
				for (int i = 0; i < audioDouble.length; i++) {
					audioData[i] = audioDouble[i];
				}
			} else {
				for (int i = 0; i < audioDouble.length; i++) {
					audioData[i] = audioDouble[i] - 128;
				}
			}
		}

		this.audioData = audioData;

		frames_per_pixel = (audioDouble.length / wavFormat.getFrameSize()) / w;

		byte my_byte = 0;
		double y_last = 0, x_last = 0;
		int numChannels = wavFormat.getChannels();

		for (double x = 0; x < w && audioData != null; x++) {
			int idx = frames_per_pixel * numChannels * (int) x;
			double temp = 0.0;
			
			try {
				temp = audioData[idx];
			} catch (Exception e) {
				e.printStackTrace();
				return;
			}

			for (int i = idx; i < idx + frames_per_pixel * numChannels; i++) {
				if ((int) x % 2 == 0) {
					if (temp > audioData[i]) {
						temp = audioData[i];
					}

				} else {
					if (temp < audioData[i]) {
						temp = audioData[i];
					}
				}

			}

			if (wavFormat.getSampleSizeInBits() == 8) {
				my_byte = (byte) temp;
			} else {
				my_byte = (byte) (128 * temp / 32768);
			}
			double y_new = (double) (h * (128 - my_byte) / 256);
			lines.add(new Line2D.Double(x_last, y_last, x, y_new));
			y_last = y_new;
			x_last = x;
		}

		repaint();
	}

	public void createWaveForm(double[] audioDouble) {
		if (wavStream == null) {
			repaint();
			return;
		}
		wavFormat = wavStream.getFormat();

		doCreateWaveForm(audioDouble, wavFormat);
	}

	public void paint(Graphics g) {
		Dimension d = getSize();
		int w = d.width;
		int h = d.height;

		Graphics2D g2 = (Graphics2D) g;
		g2.setBackground(getBackground());
		g2.clearRect(0, 0, w, h);

		for (int i = 0; i < 10; i++) {
			Shape l = new Line2D.Float(w / 10 * i, 0, w / 10 * i, h);
			g2.setColor(Color.red);
			g2.draw(l);
		}

		for (int i = 0; i < 10; i++) {
			Shape l = new Line2D.Float(w / 10 * i, 0, w / 10 * i, h);
			g2.setColor(Color.GRAY);
			g2.draw(l);
			l = new Line2D.Float(0, h / 10 * i, w, h / 10 * i);
			g2.draw(l);
		}

		Shape l = new Line2D.Float(0, h / 2, w, h / 2);
		if (!initialized) {
			g2.setColor(Color.red);
			g2.draw(l);
		}

		if (initialized) {
			g2.setColor(Color.red);
			g2.draw(l);
			g2.setColor(Color.CYAN);
			for (int i = 1; i < lines.size(); i++) {
				g2.draw((Line2D) lines.get(i));
			}
		}
	}

	public void start() {
		thread = new Thread(this);
		thread.setName("SamplingGraph");
		thread.start();
	}

	public void stop() {
		if (thread != null) {
			thread.interrupt();
		}
		thread = null;
	}

	@SuppressWarnings("static-access")
	public void run() {
		while (thread != null) {
			try {
				thread.sleep(100);
			} catch (Exception e) {
				break;
			}
			repaint();
		}
	}
}